/*
    *
    * This file is a part of CoreImage.
    * An image viewer for CuboCore Application Suite
    * Copyright 2019 CuboCore Group
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 2 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, write to the Free Software
    * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    * MA 02110-1301, USA.
    *
*/

#include <QGraphicsScene>
#include <QListWidgetItem>
#include <QFileDialog>
#include <QStandardPaths>
#include <QImageReader>
#include <QImageWriter>
#include <QFileInfo>
#include <QPicture>
#include <QSpacerItem>
#include <QtConcurrent>
#include <QShortcut>
#include <QCloseEvent>
#include <QMessageBox>

#include <cprime/themefunc.h>
#include <cprime/cprime.h>
#include <cprime/appopenfunc.h>
#include <cprime/pinit.h>
#include <cprime/trashmanager.h>
#include <cprime/shareit.h>

#include "coreimage.h"
#include "ui_coreimage.h"

QStringList coreimage::openFilters = QStringList();
QStringList coreimage::saveFilters = QStringList();


coreimage::coreimage(QWidget *parent) : QWidget(parent), ui(new Ui::coreimage), slideShowTimer(nullptr)
{
    ui->setupUi(this);

    QPalette pltt(palette());
    pltt.setColor(QPalette::Base, Qt::transparent);
    setPalette(pltt);

	imagesList.clear();

	workFilePath = nullptr;

	loadSettings();
	startMode();
	startSetup();
	shortcuts();

	getImageFilters();

	slideShow = false;
}

coreimage::~coreimage()
{
    if (slideShowTimer) {
		delete slideShowTimer;
    }

	delete ui;
}

/**
 * @brief Setup ui elements
 */
void coreimage::startSetup()
{
	zoomLbl = new QLabel();
	zoomLbl->setStyleSheet( "QLabel{ margin: 5px; padding: 10px; background-color: rgba(0, 0, 0, 0.5); border-radius: 3px; }" );
	zoomLbl->setText( "Zoom: 1.0x" );

	ui->gridLayout->addWidget( zoomLbl, 1, 2, 1, 1, Qt::AlignLeft | Qt::AlignTop );

    // all toolbuttons icon size in sideView
    for (QToolButton *b : ui->sideView->findChildren<QToolButton *>()) {
        if (b) {
            b->setIconSize(sideViewIconSize);
        }
    }

    // all toolbuttons icon size in toolBar
    for (QToolButton *b : ui->toolBar->findChildren<QToolButton *>()) {
        if (b) {
            b->setIconSize(toolBarIconSize);
        }
    }

	// set stylesheet
    ui->sideView->setStyleSheet(CPrime::ThemeFunc::getSideViewStyleSheet());
    ui->sideView->setFixedWidth(sideViewSize);

	// set window size
    if (touch) {
        this->setWindowState(Qt::WindowMaximized);
    } else {
        this->resize(800, 500);
    }

    ui->page->setCurrentIndex(0);

    connect(ui->viewWidget, &view::zoomChanged, [this]() {
			zoomLbl->setText( "Zoom: " + QString::number( ui->viewWidget->zoomFactor(), 'f', 2 ) + "x" );
		}
    );

    connect(ui->cOpen, SIGNAL(clicked()), this, SLOT(openFileDialog()));
    connect(ui->addFiles, SIGNAL(clicked()), this, SLOT(openFileDialog()));
    connect(ui->viewWidget, SIGNAL(nextImage()), this, SLOT(nextItem()));
    connect(ui->viewWidget, SIGNAL(prevImage()), this, SLOT(previousItem()));
    connect(ui->cNext, SIGNAL(clicked()), this, SLOT(nextItem()));
    connect(ui->cPrevious, SIGNAL(clicked()), this, SLOT(previousItem()));
    connect(ui->cProperties, SIGNAL(clicked()), this, SLOT(showProperties()));
    connect(ui->cZoomIn, SIGNAL(clicked()), ui->viewWidget, SLOT(zoomIn()));
    connect(ui->cZoomOut, SIGNAL(clicked()), ui->viewWidget, SLOT(zoomOut()));
    connect(ui->cNormalSize, SIGNAL(clicked()), ui->viewWidget, SLOT(zoomNormal()));
    connect(ui->tools, SIGNAL(clicked()), this, SLOT(showSideView()));
    connect(ui->appTitle, SIGNAL(clicked()), this, SLOT(showSideView()));
}

/**
 * @brief Loads application settings
 */
void coreimage::loadSettings()
{
    sideViewSize = sm->value("CoreApps", "SideViewSize");
    sideViewIconSize = sm->value("CoreApps", "SideViewIconSize");
    toolBarIconSize = sm->value("CoreApps", "ToolBarIconSize");
    touch = sm->value("CoreApps", "TouchMode");
}

void coreimage::shortcuts()
{
	QShortcut *shortcut;
    shortcut = new QShortcut(QKeySequence(Qt::Key_Left), this);
    connect(shortcut, SIGNAL(activated()), this, SLOT(previousItem()));

    shortcut = new QShortcut(QKeySequence(Qt::Key_Right), this);
    connect(shortcut, &QShortcut::activated, this, &coreimage::nextItem);

    shortcut = new QShortcut(QKeySequence(Qt::ControlModifier + Qt::Key_O), this);
    connect(shortcut, &QShortcut::activated, this, &coreimage::openFileDialog);

    shortcut = new QShortcut(QKeySequence(Qt::ControlModifier + Qt::Key_S), this);
    connect(shortcut, &QShortcut::activated, this, &coreimage::on_cSave_clicked);

    shortcut = new QShortcut(QKeySequence(Qt::ControlModifier + Qt::SHIFT + Qt::Key_S), this);
    connect(shortcut, &QShortcut::activated, this, &coreimage::on_cSaveAs_clicked);

    shortcut = new QShortcut(QKeySequence(Qt::ControlModifier + Qt::Key_D), this);
    connect(shortcut, &QShortcut::activated, this, &coreimage::on_cTrashIt_clicked);

    shortcut = new QShortcut(QKeySequence(Qt::ControlModifier + Qt::Key_Plus), this);
    connect(shortcut, &QShortcut::activated,  ui->viewWidget, &view::zoomIn);

    shortcut = new QShortcut(QKeySequence(Qt::ControlModifier + Qt::Key_Minus), this);
    connect(shortcut, &QShortcut::activated,  ui->viewWidget, &view::zoomOut);

    shortcut = new QShortcut(QKeySequence(Qt::ControlModifier + Qt::Key_0), this);
    connect(shortcut, &QShortcut::activated,  ui->viewWidget, &view::zoomNormal);

    shortcut = new QShortcut(QKeySequence(Qt::ControlModifier + Qt::Key_B), this);
    connect(shortcut, &QShortcut::activated, this, &coreimage::on_pinIt_clicked);

    shortcut = new QShortcut(QKeySequence(Qt::Key_Escape), this);
    connect(shortcut, &QShortcut::activated, [this]() {
        if(slideShow){
            on_slideShow_clicked(false);
            ui->slideShow->setChecked(false);
        }
    });
}

void coreimage::sendFiles(const QStringList &paths)
{
    if (paths.count()) {
        QString str = paths.at(0);
        if (CPrime::FileUtils::isFile(str)) {
            loadFile(str);
        } else {
            qDebug() << "func(sendFiles) : Error : Invalid file path sent :" << str;
        }
	}
}

void coreimage::closeEvent(QCloseEvent *event)
{
	event->ignore();
    CPrime::InfoFunc::saveToRecent("coreimage", workFilePath);
	event->accept();
}

void coreimage::startMode()
{
    for (QToolButton *b : ui->toolBar->findChildren<QToolButton *>()) {
        b->setEnabled(false);
	}

    for (QToolButton *b : ui->sideView->findChildren<QToolButton *>()) {
        b->setEnabled(false);
    }

    ui->cOpen->setEnabled(true);
    ui->tools->setEnabled(true);
    ui->sideView->setVisible(false);
    ui->propertisesL->setVisible(false);
}

void coreimage::getImages(const QString &path)
{
	imagesList.clear();
    QDir dir(path);

    for (QFileInfo info : dir.entryInfoList(QDir::Files | QDir::NoDotAndDotDot | QDir::Readable)) {
        if (QImageReader::supportedImageFormats().contains(info.suffix().toLocal8Bit().toLower())) {
            imagesList.append(info.filePath());
        }
    }

    imagesList = CPrime::SortFunc::sortList(imagesList, CPrime::SortOrder::Descending);
}

static void initializeImageFileDialog(QFileDialog &dialog, QFileDialog::AcceptMode acceptMode)
{
    static bool firstDialog = true;

    if (firstDialog) {
        firstDialog = false;
        const QStringList picturesLocations = QStandardPaths::standardLocations(QStandardPaths::PicturesLocation);
        dialog.setDirectory(picturesLocations.isEmpty() ? QDir::currentPath() : picturesLocations.last());
    }

    dialog.setAcceptMode(acceptMode);

    // Mimetype collection
    QStringList mimeTypeFilters;
    QByteArrayList imageFormats = QImageReader::supportedImageFormats();
    QMimeDatabase mimeDB;
    QString allType = "All Images (";

    foreach (QMimeType mimeType, mimeDB.allMimeTypes()) {
        if (mimeType.name().left(5) == "image") {
            if (imageFormats.contains(mimeType.preferredSuffix().toLocal8Bit())) {
                if (mimeType.preferredSuffix().isEmpty()) {
                    continue;
                }

                mimeTypeFilters.append(mimeType.filterString());

                foreach (QString format, mimeType.suffixes()) {
                    allType += "*." + format + " ";
                }
            }
        }
    }

    allType += ")";
    mimeTypeFilters.sort();
    mimeTypeFilters.insert(0, allType);
    dialog.setNameFilters(mimeTypeFilters);
    // ===================

    if (acceptMode == QFileDialog::AcceptSave) {
        dialog.setDefaultSuffix(dialog.selectedNameFilter());
    }
}

bool coreimage::loadFile(const QString &fileName)
{
    if (!fileName.count()) {
        CPrime::InfoFunc::messageEngine("CoreImage", "coreimage", "Empty file name!!!", "Select a valid file", this);
		return false;
	}

    ui->page->setCurrentIndex(1);

    bool ok = ui->viewWidget->loadImage(fileName);

    if (ok) {
        for (QToolButton *b : ui->toolBar->findChildren<QToolButton *>()) {
            b->setEnabled(true);
        }

        for (QToolButton *b : ui->sideView->findChildren<QToolButton *>()) {
            b->setEnabled(true);
        }

		workFilePath = fileName;
        setImage(ui->viewWidget->asImage());
    } else {
        QString mess = tr("Could not load image <b>%1</b>").arg(QFileInfo(fileName).fileName()) ;
        CPrime::InfoFunc::messageEngine("CoreImage", "coreimage", "Failed to load", mess, this);
		return false;
	}

	return ok;
}

void coreimage::setImage(const QImage &newImage)
{
    getImages(CPrime::FileUtils::dirName(workFilePath));

    QFileInfo info(workFilePath);
    setWindowTitle(info.fileName() + " - CoreImage");


    if(ui->propertisesL->isVisible()){
        QString n = info.fileName();
        QString h = QString::number(newImage.height()) ;
        QString w = QString::number(newImage.width()) ;
        QString s = CPrime::FileFunc::formatSize(static_cast<quint64>(info.size()));
        QString typ = info.suffix();
        ui->propertisesL->setText(QString("%1; %2 px x %3 px; %4; %5;")
                                  .arg(n).arg(h).arg(w).arg(s).arg(typ));
    }
}

void coreimage::on_cSave_clicked()
{
    if (ui->viewWidget->saveImage()) {
        CPrime::InfoFunc::messageEngine("CoreImage", "coreimage", "Image Saved", "Check the destination for the file.", this);
	}
}

void coreimage::on_cSaveAs_clicked()
{
    QFileDialog dialog(this, tr("Save File As"));
    initializeImageFileDialog(dialog, QFileDialog::AcceptSave);

    while (dialog.exec() == QDialog::Accepted && !ui->viewWidget->saveImage(dialog.selectedFiles().first())) {}
}

void coreimage::previousItem()
{
    if (imagesList.count() == 1) {
        return;
    }

    int index = imagesList.indexOf(workFilePath);
    int newIndex = index + 1;

    /* Loop through all images, so that we can load a proper one */
    while (newIndex != index) {
        workFilePath = imagesList.value(newIndex);
        if (workFilePath.count() == 0 || not loadFile(imagesList.value(newIndex))) {
            newIndex++;

            if (newIndex >= imagesList.count()) {
                newIndex = 0;
            }
        } else {
            index = newIndex;
            break;
        }
    }
}

void coreimage::nextItem()
{
    if (imagesList.count() == 1) {
        return;
    }

    int index = imagesList.indexOf(workFilePath);
    int newIndex = index - 1;

    /* Loop through all images, so that we can load a proper one */
    while (newIndex != index) {
        workFilePath = imagesList.value(newIndex);
        if (workFilePath.count() == 0 || not loadFile(imagesList.value(newIndex))) {
            newIndex--;

            if (newIndex < 0) {
                newIndex = imagesList.count() - 1;
            }
        } else {
            index = newIndex;
            break;
        }
    }
}

void coreimage::showProperties()
{
    // unchecked the tools button
    if (ui->tools->isChecked()) {
        ui->tools->setChecked(false);
    }

    if (!ui->propertisesL->isVisible()) {
        ui->propertisesL->setVisible(true);
        ui->sideView->setVisible(false);

        // load the current image in the view again
        loadFile(workFilePath);
	}

    else {
        ui->propertisesL->setVisible(false);
    }
}

void coreimage::on_cTrashIt_clicked()
{
    int index = imagesList.indexOf(workFilePath);

	// Function from utilities.cpp
    if (CPrime::TrashManager::moveToTrash(QStringList() << workFilePath) == true) {
        imagesList.removeAt(index);

        if (imagesList.count() == 0) {
            ui->viewWidget->scene()->clear();
            ui->propertisesL->clear();
            ui->viewWidget->clear();
            this->setWindowTitle("CoreImage");
			startMode();
        } else if (imagesList.count() > 0) {
            if (((index < (imagesList.count())) && index > 0)) {
                workFilePath = imagesList.at(index);
				nextItem();
            } else if (index == (imagesList.count())) {
                workFilePath = imagesList.at(index - 1);
				previousItem();
            } else if (index == 0) {
                workFilePath = imagesList.at(0);
                loadFile(workFilePath);
            }
		}
	}
}

void coreimage::on_containingFolder_clicked()
{
    CPrime::AppOpenFunc::defaultAppEngine(CPrime::Category::FileManager, workFilePath, "", this);
}

void coreimage::on_openInEditor_clicked()
{
    CPrime::AppOpenFunc::defaultAppEngine(CPrime::Category::ImageEditor, workFilePath, "", this);
}

void coreimage::on_openmetadataApp_clicked()
{
    CPrime::AppOpenFunc::defaultAppEngine(CPrime::Category::MetadataViewer, workFilePath, "", this);
}

void coreimage::on_slideShow_clicked(bool checked)
{
    if (checked) {
		ui->viewWidget->zoomFit();

        if (!slideShowTimer) {
			slideShowTimer = new QTimer();
			// switch to the next image when timeout
            connect(slideShowTimer, &QTimer::timeout, this, &coreimage::nextItem);
		}

        slideShowTimer->start(3000);
        ui->sideView->setVisible(false);

        if (ui->tools->isChecked()) {
            ui->tools->setChecked(false);
        }

        slideShow = true;
	}

	else {
        if (slideShowTimer) {
			delete slideShowTimer;
			slideShowTimer = nullptr;
            slideShow = false;
		}
	}

    this->setWindowState(this->windowState() ^ Qt::WindowFullScreen);
}

void coreimage::on_pinIt_clicked()
{
    if (!workFilePath.isNull()) {
		pinit pit;
        pit.callPinIt(this, workFilePath);
	}
}

void coreimage::getImageFilters()
{
	openFilters.clear();
	saveFilters.clear();

	/* Open Filters */
	openFilters << "All Images (";

    Q_FOREACH (QByteArray type, QImageReader::supportedImageFormats()) {
        openFilters[ 0 ] += "*." + QString(type) + " ";
        openFilters << QString("%1 File (*.").arg(QString(type.toUpper())) + type + ")";
    }

	openFilters[ 0 ] += ")";

	/* Save Filters */
	saveFilters << "All Images (";

    Q_FOREACH (QByteArray type, QImageWriter::supportedImageFormats()) {
        saveFilters[ 0 ] += "*." + QString(type) + " ";
        saveFilters << QString("%1 File (*.").arg(QString(type.toUpper())) + type + ")";
    }

	saveFilters[ 0 ] += ")";
}

void coreimage::on_shearIt_clicked()
{
    if (!workFilePath.isNull()) {
        shareit::shareFiles( this, QStringList() << workFilePath );
    }
}

void coreimage::showSideView()
{
    // unchecked the properties button
    if (ui->cProperties->isChecked()) {
        ui->cProperties->setChecked(false);
    }

    if (ui->sideView->isVisible()) {
        ui->sideView->setVisible(false);

    } else {
        ui->sideView->setVisible(true);

        if (ui->propertisesL->isVisible()) {
            ui->propertisesL->setVisible(false);
        }
    }
}

void coreimage::openFileDialog()
{
    QFileDialog dialog(this, tr("Open File"));
    initializeImageFileDialog(dialog, QFileDialog::AcceptOpen);

    while (dialog.exec() == QDialog::Accepted && !loadFile(dialog.selectedFiles().first())) { }
}

void coreimage::on_rotate_clicked()
{
    ui->viewWidget->rotateRight();
}
