/*
    *
    * This file is a part of CoreImage.
    * An image viewer for CuboCore Application Suite
    * Copyright 2019 CuboCore Group
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 2 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, write to the Free Software
    * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    * MA 02110-1301, USA.
    *
*/

#include "view.h"

#include <QWheelEvent>

view::view(QWidget *parent) : QGraphicsView(parent)
{
    setRenderHints(QPainter::HighQualityAntialiasing | QPainter::Antialiasing | QPainter::TextAntialiasing);
    setCacheMode(QGraphicsView::CacheBackground);
    setViewportUpdateMode(QGraphicsView::BoundingRectViewportUpdate);

    image = new QGraphicsPixmapItem();
    image->setTransformationMode(Qt::SmoothTransformation);

	mScene = new QGraphicsScene();
    mScene->addItem(image);

    setScene(mScene);
}

bool view::loadImage(QString fileName)
{
	mFileName = fileName;

    QPixmap pix(fileName);

    if (pix.isNull()) {
		return false;
    }

    image->setPixmap(pix);
    mScene->setSceneRect(image->boundingRect());
	zoomFit();

    return true;
}

void view::clear()
{
    isFit = false;
}

QImage view::asImage()
{
	return image->pixmap().toImage();
}

qreal view::zoomFactor() {

	return mScale;
}

void view::zoomIn()
{
    scale(scaleFactor, scaleFactor);
	isFit = false;

	mScale *= scaleFactor;
	zoomChanged();
}

void view::zoomOut()
{
    scale(invScaleFactor, invScaleFactor);
	isFit = false;

	mScale *= invScaleFactor;
	zoomChanged();
}

void view::zoomNormal()
{
	resetMatrix();
	isFit = false;
	mScale = 1.0;
	zoomChanged();
}

void view::zoomFit()
{
    fitInView(image, Qt::KeepAspectRatio);
	isFit = true;

	QMatrix mtx = matrix();
	mScale = mtx.m11();

	zoomChanged();
}

void view::rotateLeft()
{
	QTransform tran;
    tran.rotate(-90);

    mScene->removeItem(image);
    image->setPixmap(image->pixmap().transformed(tran, Qt::SmoothTransformation));
    mScene->addItem(image);
    mScene->setSceneRect(image->boundingRect());

    if (isFit) {
		zoomFit();
    }
}

void view::rotateRight()
{
	QTransform tran;
    tran.rotate(90);

    mScene->removeItem(image);
    image->setPixmap(image->pixmap().transformed(tran, Qt::SmoothTransformation));
    mScene->addItem(image);
    mScene->setSceneRect(image->boundingRect());

    if (isFit) {
		zoomFit();
    }
}

bool view::saveImage(QString fileName)
{
    if (fileName.count()) {
		mFileName = fileName;
    }

    return image->pixmap().save(mFileName);
}

bool view::eventFilter(QObject *, QEvent *event)
{
    if (event->type() == QEvent::Wheel) {
        QWheelEvent *wheel_event = static_cast<QWheelEvent *>(event);

        if (Qt::RightButton & wheel_event->buttons()) {   // right button + wheel
            if (wheel_event->delta() > 0) {
                zoomIn();
            } else if (wheel_event->delta() < 0) {
                zoomOut();
            }
        } else if (wheel_event->modifiers() == Qt::ControlModifier) {   // ctrl + wheel
            if (wheel_event->delta() > 0) {
                zoomIn();
            } else {
                zoomOut();
            }
        } else if (wheel_event->buttons() == 0) {  // only wheel
            if (wheel_event->delta() < 0) {
				emit nextImage();
            } else if (wheel_event->delta() > 0) {
				emit prevImage();
            }
        }
	}

	return false;
}

void view::mouseDoubleClickEvent(QMouseEvent *event)
{
    if (event->button() == Qt::LeftButton) {
        if (isFit) {
			zoomNormal();
        } else {
			zoomFit();
        }
    }
}

void view::resizeEvent(QResizeEvent *rEvent)
{
	rEvent->accept();

    if (isFit) {
		zoomFit();
    }
}
